/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2012 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.graphicsGL.ui.util;

import static org.fortiss.tooling.graphicsGL.ui.util.FloatBufferUtils.toFloatBuffer;
import static org.fortiss.tooling.graphicsGL.util.Mesh3DModelElementFactory.DEFAULT_MATERIAL;
import static org.lwjgl.BufferUtils.createFloatBuffer;
import static org.lwjgl.opengl.GL11.GL_AMBIENT;
import static org.lwjgl.opengl.GL11.GL_AMBIENT_AND_DIFFUSE;
import static org.lwjgl.opengl.GL11.GL_BACK;
import static org.lwjgl.opengl.GL11.GL_BLEND;
import static org.lwjgl.opengl.GL11.GL_COLOR_MATERIAL;
import static org.lwjgl.opengl.GL11.GL_CONSTANT_ATTENUATION;
import static org.lwjgl.opengl.GL11.GL_CULL_FACE;
import static org.lwjgl.opengl.GL11.GL_DEPTH_TEST;
import static org.lwjgl.opengl.GL11.GL_DIFFUSE;
import static org.lwjgl.opengl.GL11.GL_FLAT;
import static org.lwjgl.opengl.GL11.GL_FRONT;
import static org.lwjgl.opengl.GL11.GL_LIGHTING;
import static org.lwjgl.opengl.GL11.GL_LIGHT_MODEL_AMBIENT;
import static org.lwjgl.opengl.GL11.GL_LINEAR_ATTENUATION;
import static org.lwjgl.opengl.GL11.GL_LINES;
import static org.lwjgl.opengl.GL11.GL_MODELVIEW;
import static org.lwjgl.opengl.GL11.GL_MODELVIEW_MATRIX;
import static org.lwjgl.opengl.GL11.GL_NICEST;
import static org.lwjgl.opengl.GL11.GL_NORMALIZE;
import static org.lwjgl.opengl.GL11.GL_ONE_MINUS_SRC_ALPHA;
import static org.lwjgl.opengl.GL11.GL_PERSPECTIVE_CORRECTION_HINT;
import static org.lwjgl.opengl.GL11.GL_POLYGON_SMOOTH;
import static org.lwjgl.opengl.GL11.GL_POLYGON_SMOOTH_HINT;
import static org.lwjgl.opengl.GL11.GL_POSITION;
import static org.lwjgl.opengl.GL11.GL_PROJECTION;
import static org.lwjgl.opengl.GL11.GL_QUADRATIC_ATTENUATION;
import static org.lwjgl.opengl.GL11.GL_RENDER;
import static org.lwjgl.opengl.GL11.GL_SELECT;
import static org.lwjgl.opengl.GL11.GL_SHININESS;
import static org.lwjgl.opengl.GL11.GL_SMOOTH;
import static org.lwjgl.opengl.GL11.GL_SPECULAR;
import static org.lwjgl.opengl.GL11.GL_SRC_ALPHA;
import static org.lwjgl.opengl.GL11.GL_TEXTURE;
import static org.lwjgl.opengl.GL11.GL_TEXTURE_2D;
import static org.lwjgl.opengl.GL11.GL_TRIANGLES;
import static org.lwjgl.opengl.GL11.GL_VIEWPORT;
import static org.lwjgl.opengl.GL11.glBegin;
import static org.lwjgl.opengl.GL11.glBlendFunc;
import static org.lwjgl.opengl.GL11.glClear;
import static org.lwjgl.opengl.GL11.glClearColor;
import static org.lwjgl.opengl.GL11.glClearDepth;
import static org.lwjgl.opengl.GL11.glColor4f;
import static org.lwjgl.opengl.GL11.glColorMaterial;
import static org.lwjgl.opengl.GL11.glCullFace;
import static org.lwjgl.opengl.GL11.glDisable;
import static org.lwjgl.opengl.GL11.glEnable;
import static org.lwjgl.opengl.GL11.glEnd;
import static org.lwjgl.opengl.GL11.glGetFloat;
import static org.lwjgl.opengl.GL11.glGetInteger;
import static org.lwjgl.opengl.GL11.glHint;
import static org.lwjgl.opengl.GL11.glInitNames;
import static org.lwjgl.opengl.GL11.glLight;
import static org.lwjgl.opengl.GL11.glLightModel;
import static org.lwjgl.opengl.GL11.glLightf;
import static org.lwjgl.opengl.GL11.glLineWidth;
import static org.lwjgl.opengl.GL11.glLoadIdentity;
import static org.lwjgl.opengl.GL11.glMaterial;
import static org.lwjgl.opengl.GL11.glMaterialf;
import static org.lwjgl.opengl.GL11.glMatrixMode;
import static org.lwjgl.opengl.GL11.glMultMatrix;
import static org.lwjgl.opengl.GL11.glNormal3f;
import static org.lwjgl.opengl.GL11.glOrtho;
import static org.lwjgl.opengl.GL11.glPopMatrix;
import static org.lwjgl.opengl.GL11.glPopName;
import static org.lwjgl.opengl.GL11.glPushMatrix;
import static org.lwjgl.opengl.GL11.glPushName;
import static org.lwjgl.opengl.GL11.glRenderMode;
import static org.lwjgl.opengl.GL11.glRotatef;
import static org.lwjgl.opengl.GL11.glScalef;
import static org.lwjgl.opengl.GL11.glSelectBuffer;
import static org.lwjgl.opengl.GL11.glShadeModel;
import static org.lwjgl.opengl.GL11.glTranslatef;
import static org.lwjgl.opengl.GL11.glVertex3f;
import static org.lwjgl.opengl.GL11.glViewport;
import static org.lwjgl.util.glu.GLU.gluLookAt;
import static org.lwjgl.util.glu.GLU.gluPerspective;
import static org.lwjgl.util.glu.GLU.gluPickMatrix;

import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.List;

import org.fortiss.tooling.graphicsGL.model.Color3D;
import org.fortiss.tooling.graphicsGL.model.Vector3D;
import org.fortiss.tooling.graphicsGL.model.mesh.Face3D;
import org.fortiss.tooling.graphicsGL.model.mesh.Material3D;
import org.fortiss.tooling.graphicsGL.model.mesh.MaterialGroup3D;
import org.fortiss.tooling.graphicsGL.model.mesh.Normal3D;
import org.fortiss.tooling.graphicsGL.model.mesh.Vertex3D;
import org.fortiss.tooling.graphicsGL.model.scene.Light3D;
import org.fortiss.tooling.graphicsGL.ui.camera.Camera;
import org.fortiss.tooling.graphicsGL.ui.objects.ViewObjectBase;
import org.fortiss.tooling.graphicsGL.ui.picker.ModelObjectPicker;
import org.lwjgl.opengl.GL11;
import org.lwjgl.util.vector.Vector3f;
import org.lwjgl.util.vector.Vector4f;
import org.newdawn.slick.Color;

/**
 * Utility class for GL primitives like setting colors, drawing lines, triangles, quads, etc.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 25892D6CA7A35E798D1B71AAEC81A964
 */
public class GLPrimitives {
	/** The X unit vector. */
	public static Vector3f X_UNIT_VECTOR = new Vector3f(1.0f, 0, 0);
	/** The Y unit vector. */
	public static Vector3f Y_UNIT_VECTOR = new Vector3f(0, 1.0f, 0);
	/** The Z unit vector. */
	public static Vector3f Z_UNIT_VECTOR = new Vector3f(0, 0, 1.0f);

	/** The GL_LIGHTx array. */
	public static final int[] GL_LIGHT_NUMBER = new int[] {GL11.GL_LIGHT0, GL11.GL_LIGHT1,
			GL11.GL_LIGHT2, GL11.GL_LIGHT3, GL11.GL_LIGHT4, GL11.GL_LIGHT5, GL11.GL_LIGHT6,
			GL11.GL_LIGHT7};

	/** Initializes basic GL properties for 3D rendering. */
	public static void initializeGLfor3D(int width, int height, float fieldOfViewAngle,
			float aspect, float nearDistance, float farDistance) {
		glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
		glClearDepth(1.0);
		glLineWidth(2);
		glEnable(GL_DEPTH_TEST);
		glDisable(GL_BLEND);

		// set viewport and alter set projection matrix
		glViewport(0, 0, width, height);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(fieldOfViewAngle, aspect, nearDistance, farDistance);
		// return GL state to model matrix
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		// enable anti aliasing
		GL11.glEnable(GL11.GL_BLEND);
		glEnable(GL_POLYGON_SMOOTH);
		glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
	}

	/** Initializes GL properties for 2D rendering. */
	public static void initializeGLfor2D(int width, int height) {
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_LIGHTING);
		glLineWidth(2);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		// orthographic projection
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(0, width, height, 0, 1, -1);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

	}

	/**
	 * Start GL picking mode rendering for the given scene, perspective, and pick location and
	 * object picker. This method returns the hits counted during picking render run. The results of
	 * the picking run are stored.
	 */
	public static int startPickingModeAndCountHits(int x, int y, Vector3f perspective,
			IntBuffer viewport, ViewObjectBase scene, ModelObjectPicker picker,
			IntBuffer selectBuffer, Camera c) {
		glGetInteger(GL_VIEWPORT, viewport);
		float aspect =
				(float)(viewport.get(2) - viewport.get(0)) /
						(float)(viewport.get(3) - viewport.get(1));

		glSelectBuffer(selectBuffer);
		glRenderMode(GL_SELECT);
		glInitNames();

		// setup perspective
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		gluPickMatrix(x, viewport.get(3) - y, 5.0f, 5.0f, viewport);
		gluPerspective(perspective.x, aspect, perspective.y, perspective.z);

		// render objects in picking mode
		glMatrixMode(GL_MODELVIEW);
		glDisable(GL_TEXTURE);
		glDisable(GL_LIGHTING);
		scene.renderObjectForPicking(picker, c);

		// restore projection matrix
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);

		// return to render mode to count hits
		int hits = glRenderMode(GL_RENDER);
		return hits;
	}

	/** Returns the current transformed origin in model view coordinates. */
	public static Vector3f getTransformedOrigin() {
		FloatBuffer fb = createFloatBuffer(16);
		glGetFloat(GL_MODELVIEW_MATRIX, fb);
		return new Vector3f(fb.get(12), fb.get(13), fb.get(14));
	}

	/** Returns the given vector in model view coordinates. */
	public static Vector3f getTransformedVector(Vector3D v) {
		FloatBuffer fb = createFloatBuffer(16);
		glGetFloat(GL_MODELVIEW_MATRIX, fb);
		float x = fb.get(0) * v.getX() + fb.get(4) * v.getY() + fb.get(8) * v.getZ() + fb.get(12);
		float y = fb.get(1) * v.getX() + fb.get(5) * v.getY() + fb.get(9) * v.getZ() + fb.get(13);
		float z = fb.get(2) * v.getX() + fb.get(6) * v.getY() + fb.get(10) * v.getZ() + fb.get(14);
		return new Vector3f(x, y, z);
	}

	/** Sets the current color. */
	public static void setColor(Color color) {
		glColor4f(color.r, color.g, color.b, 1.0f);
	}

	/** Sets the current color. */
	public static void setColor(Color3D color) {
		glColor4f(color.getRed(), color.getGreen(), color.getBlue(), 1.0f);
	}

	/** Applies the given translation. */
	public static void translate(Vector3f t) {
		glTranslatef(t.getX(), t.getY(), t.getZ());
	}

	/** Applies the given translation. */
	public static void translate(float x, float y, float z) {
		glTranslatef(x, y, z);
	}

	/** Applies the given scaling factor. */
	public static void scale(Vector3f scale) {
		glScalef(scale.x, scale.y, scale.z);
	}

	/** Applies the given scaling factor. */
	public static void scale(float x, float y, float z) {
		glScalef(x, y, z);
	}

	/** Applies the given rotation angle to the axis. */
	public static void rotate(float f, float x, float y, float z) {
		glRotatef(f, x, y, z);
	}

	/** Applies the given rotation angle to the axis. */
	public static void rotate(Vector4f axisAndAngle) {
		glRotatef(axisAndAngle.w, axisAndAngle.x, axisAndAngle.y, axisAndAngle.z);
	}

	/** Applies the given rotation angle to the given axis. */
	public static void rotate(float angle, Vector3f axis) {
		glRotatef(angle, axis.x, axis.y, axis.z);
	}

	/** Sets as new transformation matrix. */
	public static void changeTransformationMatrix(float[] floatBuffer) {
		changetransformationMatrix(FloatBufferUtils.toFloatBuffer(floatBuffer));
	}

	/** Sets as new transformation matrix. */
	public static void changetransformationMatrix(FloatBuffer floatBuffer) {
		glMultMatrix(floatBuffer);
	}

	/** Draws a line between the two vertices. */
	public static void drawLine(Vector3f v1, Vector3f v2) {
		drawLine(v1.x, v1.y, v1.z, v2.x, v2.y, v2.z);
	}

	/** Draws a line between the two points. */
	public static void drawLine(Vector3D p1, Vector3D p2) {
		drawLine(p1.getX(), p1.getY(), p1.getZ(), p2.getX(), p2.getY(), p2.getZ());
	}

	/** Draws a line between the two points. */
	public static void drawLine(float x1, float y1, float z1, float x2, float y2, float z2) {
		glBegin(GL_LINES);
		glVertex3f(x1, y1, z1);
		glVertex3f(x2, y2, z2);
		glEnd();
	}

	/** Draws the given list of material groups. */
	public static void drawMaterialGroups(List<MaterialGroup3D> groups) {
		drawMaterialGroups(groups, true);
	}

	/**
	 * Draws the given list of material groups setting material properties only when
	 * {@code useLight} is {@code true}. Otherwise, the ambient component of the material is used to
	 * set the GL color property.
	 */
	public static void drawMaterialGroups(List<MaterialGroup3D> groups, boolean useLight) {
		for(MaterialGroup3D g : groups) {
			Material3D mat = g.getMaterial() == null ? DEFAULT_MATERIAL : g.getMaterial();
			setSmoothShading(g.isSmoothShading());
			if(useLight) {
				setMaterial(mat);
			} else {
				setColor(mat.getAmbient());
			}
			drawTriangleFaces(g.getFacesList());
		}
	}

	/** Draws the given list of {@link Face3D}s with triangle information. */
	public static void drawTriangleFaces(List<Face3D> faces) {
		glBegin(GL_TRIANGLES);
		for(Face3D f : faces) {
			boolean useNormals = f.getNormalsLength() == 3;
			if(!useNormals) {
				System.out.println("normals length = " + f.getNormalsLength());
			}
			for(int i = 0; i < 3; i++) {
				if(useNormals) {
					Normal3D n = f.getNormals(i);
					glNormal3f(n.getX(), n.getY(), n.getZ());
				}
				Vertex3D v = f.getVertices(i);
				glVertex3f(v.getX(), v.getY(), v.getZ());
			}
		}
		glEnd();
	}

	/** Saves the current matrix state. */
	public static void saveMatrix() {
		glPushMatrix();
	}

	/** Restores the current matrix state. */
	public static void restoreMatrix() {
		glPopMatrix();
	}

	/** Translates the matrix to the given coordinates. */
	public static void moveTo(Vector3f coords) {
		glTranslatef(coords.x, coords.y, coords.z);
	}

	/** Translates the matrix to the given coordinates. */
	public static void moveTo(Vector3D coords) {
		glTranslatef(coords.getX(), coords.getY(), coords.getZ());
	}

	/** Sets the current material data for GL_FRONT faces. */
	public static void setMaterial(FloatBuffer ambient, FloatBuffer diffuse, FloatBuffer specular,
			float specularCoeff) {
		// glDisable(GL_COLOR_MATERIAL);
		// glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		glMaterial(GL_FRONT, GL_AMBIENT, ambient);
		glMaterial(GL_FRONT, GL_DIFFUSE, diffuse);
		glMaterial(GL_FRONT, GL_SPECULAR, specular);
		glMaterialf(GL_FRONT, GL_SHININESS, specularCoeff);
	}

	/** Sets the current material data for GL_FRONT faces. */
	public static void setMaterial(Material3D mat) {
		setMaterial(toFloatBuffer(mat.getAmbient(), 1.0f), toFloatBuffer(mat.getDiffuse(), 1.0f),
				toFloatBuffer(mat.getSpecular(), 1.0f), mat.getSpecularCoeff());
	}

	/** Enables the light with the given number and properties. */
	public static void enableLight(int glLightNumber, FloatBuffer lightPosition,
			FloatBuffer ambient, FloatBuffer diffuse, FloatBuffer specular, float constAtt,
			float linAtt, float quadAtt) {
		glLight(glLightNumber, GL_POSITION, lightPosition);
		glLight(glLightNumber, GL_AMBIENT, ambient);
		glLight(glLightNumber, GL_DIFFUSE, diffuse);
		glLight(glLightNumber, GL_SPECULAR, specular);
		glLightf(glLightNumber, GL_CONSTANT_ATTENUATION, constAtt);
		glLightf(glLightNumber, GL_LINEAR_ATTENUATION, linAtt);
		glLightf(glLightNumber, GL_QUADRATIC_ATTENUATION, quadAtt);
		glEnable(glLightNumber);
	}

	/** Enables the given light with the given number and model data. */
	public static void enableLight(int glLightNumber, Light3D light) {
		enableLight(glLightNumber,
				toFloatBuffer(light.getPosition(), light.isDirectional() ? 0.0f : 1.0f),
				toFloatBuffer(light.getAmbient(), 1.0f), toFloatBuffer(light.getDiffuse(), 1.0f),
				toFloatBuffer(light.getSpecular(), 1.0f), light.getConstantAttenuation(),
				light.getLinearAttenuation(), light.getQuadraticAttenuation());
	}

	/** Clears the canvas with the given RGB color. */
	public static void clearCanvasAndLoadIdentity(Color color) {
		glMatrixMode(GL_MODELVIEW);
		glClearColor(color.r, color.g, color.b, 1.0f);
		glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT);
		loadIdentity();
	}

	/** Loads the identity matrix. */
	public static void loadIdentity() {
		glLoadIdentity();
	}

	/** Creates a transformation matrix for billboards. Result is stored in bbmat */
	public static void createBillboardMatrix(float[] bbmat, Vector3f right, Vector3f up,
			Vector3f look, Vector3f pos) {

		bbmat[0] = right.x;
		bbmat[1] = right.y;
		bbmat[2] = right.z;
		bbmat[3] = 0;
		bbmat[4] = up.x;
		bbmat[5] = up.y;
		bbmat[6] = up.z;
		bbmat[7] = 0;
		bbmat[8] = look.x;
		bbmat[9] = look.y;
		bbmat[10] = look.z;
		bbmat[11] = 0;
		// Add the translation in as well.
		bbmat[12] = pos.x;
		bbmat[13] = pos.y;
		bbmat[14] = pos.z;
		bbmat[15] = 1;

	}

	/** Prepare a name for the given object and picker. */
	public static void prepareName(Object obj, ModelObjectPicker picker) {
		int name = picker.registerPickableObject(obj);
		glPushName(name);
	}

	/** Flushes the last name. */
	public static void flushName() {
		glPopName();
	}

	/** Enables culling for back polygons (i.e. only front faces are considered for lighting. */
	public static void enableBackCulling() {
		glEnable(GL_CULL_FACE);
		glCullFace(GL_BACK);
	}

	/** Enables the normlization for provided normals. */
	public static void enableNormalization() {
		glEnable(GL_NORMALIZE);
	}

	/** Enables lighting with the given ambient intensity. */
	public static void enableLighting(Color3D col) {
		glEnable(GL_LIGHTING);
		FloatBuffer ambient = toFloatBuffer(col.getRed(), col.getGreen(), col.getBlue(), 1.0f);
		glLightModel(GL_LIGHT_MODEL_AMBIENT, ambient);
	}

	/** Enables color material for front faces and ambient and diffuse light components. */
	public static void enableColorMaterial() {
		glEnable(GL_COLOR_MATERIAL);
		glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	}

	/** Sets the smoothness shading mode. */
	public static void setSmoothShading(boolean smooth) {
		glShadeModel(smooth ? GL_SMOOTH : GL_FLAT);
	}

	/** Enables 2D textures. */
	public static void enableTextures() {
		glEnable(GL_TEXTURE_2D);
	}

	/** Disables 2D textures. */
	public static void disableTextures() {
		glDisable(GL_TEXTURE_2D);
	}

	/**
	 * Positions the camera at the given location looking at the target location. The up vector is
	 * the Z axis.
	 */
	public static void lookAt(float posX, float posY, float posZ, float targetX, float targetY,
			float targetZ) {
		gluLookAt(posX, posY, posZ, targetX, targetY, targetZ, 0.0f, 0.0f, 1.0f);
	}

}
